"""
Python part of FS_36_17_84_44 `check_env` implementation.
"""

import os
import sys
from enum import Enum, auto

from argrelay import misc_helper_common
from argrelay.check_env.CheckEnvResult import CheckEnvResult
from argrelay.check_env.PluginCheckEnvAbstract import PluginCheckEnvAbstract
from argrelay.enum_desc.PluginSide import PluginSide
from argrelay.enum_desc.PluginType import PluginType
from argrelay.enum_desc.ResultCategory import ResultCategory
from argrelay.enum_desc.TermColor import TermColor
from argrelay.relay_client.client_utils import handle_main_exception
from argrelay.runtime_context.PluginClientAbstract import instantiate_client_plugin
from argrelay.runtime_data.CheckEnvPluginConfig import CheckEnvPluginConfig
from argrelay.runtime_data.PluginEntry import PluginEntry
from argrelay.schema_config_plugin.CheckEnvPluginConfigSchema import check_env_plugin_config_desc

# Standard color scheme has to be synced with `@/exe/check_env.bash`:
success_color = f"{TermColor.back_dark_green.value}{TermColor.fore_dark_black.value}"
warning_color = f"{TermColor.back_dark_yellow.value}{TermColor.fore_dark_black.value}"
failure_color = f"{TermColor.back_dark_red.value}{TermColor.fore_bright_white.value}"
field_color = TermColor.fore_bright_cyan.value
failure_message = TermColor.fore_dark_red.value
warning_message = TermColor.fore_bright_yellow.value
success_message = TermColor.fore_bright_green.value
reset_style = TermColor.reset_style.value

# Extra color scheme (available via Python code only):
offline_color = f"{TermColor.back_light_gray.value}{TermColor.fore_dark_black.value}"
offline_message = TermColor.fore_bright_blue.value


# TODO: TODO_67_33_03_53.add_check_env_test_support.md
def main():
    # noinspection PyBroadException
    try:
        return check_env_logic()
    except BaseException as e1:
        handle_main_exception(e1)

def check_env_logic():
    argrelay_dir: str = os.path.realpath(os.path.abspath(sys.argv[1]))
    misc_helper_common.set_argrelay_dir(argrelay_dir)

    if len(sys.argv) > 3:
        raise ValueError(f"too many arguments: {' '.join(sys.argv[1:])}")
    elif len(sys.argv) == 3:
        second_arg = sys.argv[2].strip()
        if second_arg == "offline":
            online_mode = False
        elif second_arg == "online":
            online_mode = True
        else:
            raise ValueError(f"unrecognized argument: {second_arg}")
    else:
        online_mode = None

    plugin_config: CheckEnvPluginConfig = check_env_plugin_config_desc.obj_from_default_file()

    total_success: bool = True

    for plugin_instance_id in plugin_config.check_env_plugin_instance_id_activate_list:
        plugin_entry: PluginEntry = plugin_config.check_env_plugin_instance_entries[plugin_instance_id]

        if not plugin_entry.plugin_enabled:
            continue

        if (
            plugin_entry.plugin_side != PluginSide.PluginClientSideOnly
            and
            plugin_entry.plugin_side != PluginSide.PluginAnySide
        ):
            continue

        plugin_instance: PluginCheckEnvAbstract = instantiate_client_plugin(
            plugin_instance_id,
            plugin_entry,
        )
        plugin_type = plugin_instance.get_plugin_type()

        if plugin_type is PluginType.CheckEnvPlugin:
            plugin_instance: PluginCheckEnvAbstract
            plugin_instance.activate_plugin()

        class OutputCategory(Enum):
            is_failure = auto()
            is_warning = auto()
            is_offline = auto()
            is_success = auto()

        check_env_results: list[CheckEnvResult] = plugin_instance.execute_check(
            online_mode,
        )
        for check_env_result in check_env_results:
            if (
                check_env_result.result_category is ResultCategory.ExecutionFailure
                or
                check_env_result.result_category is ResultCategory.VerificationFailure
            ):
                output_category = OutputCategory.is_failure
            elif check_env_result.result_category is ResultCategory.ServerOffline:
                output_category = OutputCategory.is_offline
            elif check_env_result.result_category is ResultCategory.VerificationWarning:
                output_category = OutputCategory.is_warning
            elif check_env_result.result_category is ResultCategory.VerificationSuccess:
                output_category = OutputCategory.is_success
            else:
                raise RuntimeError(check_env_result)

            total_success = total_success and not (output_category is OutputCategory.is_failure)

            # Print level:
            if output_category is OutputCategory.is_failure:
                level_color = failure_color
                level_name = "ERROR"
            elif output_category is OutputCategory.is_offline:
                level_color = offline_color
                level_name = "SKIP"
            elif output_category is OutputCategory.is_warning:
                level_color = warning_color
                level_name = "WARN"
            else:
                level_color = success_color
                level_name = "INFO"
            print(f"{level_color}{level_name}:{reset_style}", end = " ")

            # Print field:
            if check_env_result.result_key is not None:
                print(f"{field_color}{check_env_result.result_key}:{reset_style}", end = " ")
            else:
                pass

            # Print value:
            if check_env_result.result_value is not None:
                print(f"{check_env_result.result_value}", end = " ")
            else:
                pass

            # Print message:
            if check_env_result.result_message is not None:
                if output_category is OutputCategory.is_failure:
                    message_color = failure_message
                elif output_category is OutputCategory.is_offline:
                    message_color = offline_message
                elif output_category is OutputCategory.is_warning:
                    message_color = warning_message
                else:
                    message_color = success_message
                print(f"{message_color}# {check_env_result.result_message}{reset_style}", end = " ")
            else:
                pass

            # Terminate line:
            print()

    if not total_success:
        from argrelay.enum_desc.ClientExitCode import ClientExitCode
        exit(ClientExitCode.GeneralError.value)


if __name__ == "__main__":
    main()



from __future__ import annotations

from dataclasses import dataclass, field

from argrelay.composite_forest.CompositeNode import ZeroArgNode


@dataclass
class CompositeForest:
    """
    Top-level entry object to traverse FS_33_76_82_84 composite forest.

    See also :class:`CompositeForestSchema`.
    """

    tree_roots: dict[str, ZeroArgNode] = field()



"""
"""


def format_time_to_relative(
    curr_time_ms,
    prev_time_ms,
) -> str:
    """
    Format time to relative:
    https://stackoverflow.com/a/6109105/441652
    """

    ms_per_second = 1000
    ms_per_minute = ms_per_second * 60
    ms_per_hour = ms_per_minute * 60
    ms_per_day = ms_per_hour * 24

    elapsed_time = curr_time_ms - prev_time_ms

    if elapsed_time < ms_per_minute:
        return f"{round(elapsed_time / ms_per_second)} seconds ago"
    elif elapsed_time < ms_per_hour:
        return f"{round(elapsed_time / ms_per_minute)} minutes ago"
    elif elapsed_time < ms_per_day:
        return f"{round(elapsed_time / ms_per_hour)} hours ago"
    else:
        return f"{round(elapsed_time / ms_per_day)} days ago"



from __future__ import annotations

from enum import Enum, auto
from typing import Union

from argrelay.composite_forest.AbstractNodeVisitor import AbstractNodeVisitor
from argrelay.composite_forest.CompositeForest import CompositeForest
from argrelay.composite_forest.CompositeNode import (
    InterpTreeNode,
    TreePathNode,
    FuncTreeNode,
    ZeroArgNode,
    BaseNode,
)


class TraverseDecision(Enum):
    walk_sub_tree = auto()
    skip_sub_tree = auto()


class CompositeForestWalkerAbstract(AbstractNodeVisitor):
    """
    FS_33_76_82_84 composite forest walker

    It implements traversal algorithms which extract various `CompositeInfoType`-s
    (info in a form simplified for specific purpose).

    The original intention is to extract info in the form which `DictTreeWalker` can consume -
    an intermediate phase to migrate to composite forest config while still using existing code
    (existing code relied on data structures processed via `DictTreeWalker`).
    """

    def __init__(
        self,
        composite_forest: CompositeForest,
    ):
        # main input:
        self.tree_roots: dict[str, ZeroArgNode] = composite_forest.tree_roots
        self.curr_node_id: Union[str, None] = None
        self.curr_node: Union[ZeroArgNode, InterpTreeNode, FuncTreeNode, TreePathNode, None] = None
        self.curr_path: list[str] = []

        # internal trackers:
        self.visitor_decision: TraverseDecision = TraverseDecision.walk_sub_tree

    def walk_tree_roots(
        self,
    ) -> None:
        for tree_root_id in self.tree_roots.keys():
            self._descend_and_ascend(
                self.tree_roots,
                tree_root_id,
            )

    def _walk_tree_node(
        self,
    ) -> None:
        self.visit_node(self.curr_node)
        if self.visitor_decision is TraverseDecision.walk_sub_tree:
            self._walk_sub_tree()
        elif self.visitor_decision is TraverseDecision.skip_sub_tree:
            pass
        self.leave_node(self.curr_node)

    def _walk_sub_tree(
        self,
    ):
        if self.curr_node.sub_tree is not None:
            for sub_tree_node_id in self.curr_node.sub_tree.keys():
                self._descend_and_ascend(
                    self.curr_node.sub_tree,
                    sub_tree_node_id,
                )

    def _descend_and_ascend(
        self,
        node_dict: dict[str, BaseNode],
        node_id: str,
    ) -> None:
        """
        Recursive DFS traversal of `composite_forest`.
        """
        # prepare:
        self.curr_path.append(node_id)
        prev_node = self.curr_node
        prev_node_id = self.curr_node_id
        self.curr_node_id = node_id
        # dive:
        self.curr_node = node_dict[node_id]
        self._walk_tree_node()
        # recover:
        del self.curr_path[-1]
        self.curr_node = prev_node
        self.curr_node_id = prev_node_id


class CompositeForestWalkerPrinter(CompositeForestWalkerAbstract):
    """
    Basic composite forest walker which prints node path and node type.
    """

    def __init__(
        self,
        composite_forest: CompositeForest,

    ):
        super().__init__(
            composite_forest,
        )

    def _visit_zero_arg_node(self, node: ZeroArgNode) -> None:
        self._print_info()

    def _visit_tree_path_node(self, node: TreePathNode) -> None:
        self._print_info()

    def _visit_interp_tree_node(self, node: InterpTreeNode) -> None:
        self._print_info()

    def _visit_func_tree_node(self, node: FuncTreeNode) -> None:
        self._print_info()

    def _print_info(self):
        print(" / ".join(self.curr_path), end = ": ")
        print(self.curr_node.node_type.name)
        print()
        self.visitor_decision = TraverseDecision.walk_sub_tree



from marshmallow import Schema, RAISE, fields

from argrelay.misc_helper_common.TypeDesc import TypeDesc

single_func_id_ = "single_func_id"


# TODO_40_10_18_32: add custom base to all schemas:
class AbstractJumpDelegatorConfigSchema(Schema):
    class Meta:
        unknown = RAISE
        strict = True

    single_func_id = fields.String(
        required = True,
    )


abstract_jump_delegator_config_example = {
    single_func_id_: "func_id_intercept_invocation",
}

abstract_jump_delegator_config_desc = TypeDesc(
    dict_schema = AbstractJumpDelegatorConfigSchema(),
    ref_name = AbstractJumpDelegatorConfigSchema.__name__,
    dict_example = abstract_jump_delegator_config_example,
    default_file_path = "",
)



from marshmallow import RAISE, fields

from argrelay.misc_helper_common.ObjectSchema import ObjectSchema
from argrelay.misc_helper_common.TypeDesc import TypeDesc
from argrelay.plugin_config.ConfiguratorDefaultConfig import ConfiguratorDefaultConfig

"""
Schema for :class:`ConfiguratorDefaultConfig`
"""

project_title_ = "project_title"

project_page_url_ = "project_page_url"

git_files_by_commit_id_url_prefix_ = "git_files_by_commit_id_url_prefix"

commit_id_url_prefix_ = "commit_id_url_prefix"


class ConfiguratorDefaultConfigSchema(ObjectSchema):
    class Meta:
        unknown = RAISE
        ordered = True

    model_class = ConfiguratorDefaultConfig

    project_title = fields.String(
        required = False,
        load_default = None,
    )

    project_page_url = fields.String(
        required = False,
        load_default = None,
    )

    git_files_by_commit_id_url_prefix = fields.String(
        required = False,
        load_default = None,
    )

    commit_id_url_prefix = fields.String(
        required = False,
        load_default = None,
    )


configurator_default_config_desc = TypeDesc(
    dict_schema = ConfiguratorDefaultConfigSchema(),
    ref_name = ConfiguratorDefaultConfigSchema.__name__,
    dict_example = {
        project_title_: "argrelay",
        project_page_url_: "https://argrelay.org",
        git_files_by_commit_id_url_prefix_: "https://github.com/argrelay/argrelay/tree/",
        commit_id_url_prefix_: "https://github.com/argrelay/argrelay/commit/",
    },
    default_file_path = "",
)



"""
A helper module which is import-latency-sensitive because it is used by a client (as well as a server).

For server-only helper, see `misc_helper_server`.
"""
import os
import sys

# This global variable is (supposed to be) overridden by `run_argrelay_client` and `run_argrelay_server` or
# whatever/whoever starts `argrelay` code to set to `@/` (`argrelay_dir`) according to FS_29_54_67_86 dir_structure:
_argrelay_dir = None


def set_argrelay_dir(argrelay_dir):
    global _argrelay_dir
    _argrelay_dir = argrelay_dir


def get_argrelay_dir():
    assert _argrelay_dir is not None
    return _argrelay_dir


# noinspection SpellCheckingInspection
def eprint(*args, **kwargs):
    print(*args, file = sys.stderr, **kwargs)


def get_config_path(conf_rel_path: str) -> str:
    """
    Implements FS_16_07_78_84 conf dir priority.
    """
    if "ARGRELAY_CONF_BASE_DIR" in os.environ:
        base_conf_dir = os.environ.get("ARGRELAY_CONF_BASE_DIR")
    else:
        user_dir = os.path.expanduser("~/.argrelay.conf.d/")
        if os.path.exists(user_dir):
            base_conf_dir = user_dir
        elif _argrelay_dir:
            base_conf_dir = f"{_argrelay_dir}/conf/"
        else:
            raise RuntimeError("argrelay_dir is not defined")
    return f"{base_conf_dir}/{conf_rel_path}"



from __future__ import annotations

from dataclasses import dataclass, field

from argrelay.composite_forest.CompositeForest import CompositeForest


@dataclass(frozen = True)
class ServerPluginControl:
    """
    Configures (controls) usage of plugins by the server.

    See also `ServerPluginControlSchema.py`.
    """

    first_interp_factory_id: str = field()

    composite_forest: CompositeForest = field()



from marshmallow import fields, RAISE

from argrelay.composite_forest.CompositeForest import CompositeForest
from argrelay.composite_forest.CompositeNodeSchema import CompositeNodeSchema, zero_arg_node_desc
from argrelay.misc_helper_common.ObjectSchema import ObjectSchema
from argrelay.misc_helper_common.TypeDesc import TypeDesc

tree_roots_ = "tree_roots"


class CompositeForestSchema(ObjectSchema):
    """
    Schema FS_33_76_82_84 composite forest (collection of composite tree roots).
    """

    class Meta:
        unknown = RAISE
        strict = True

    model_class = CompositeForest

    tree_roots = fields.Dict(
        keys = fields.String(),
        values = fields.Nested(CompositeNodeSchema()),
        required = True,
    )


_composite_forest_example = {
    tree_roots_: {
        "lay": zero_arg_node_desc.dict_example,
    },
}

composite_forest_desc = TypeDesc(
    dict_schema = CompositeForestSchema(),
    ref_name = CompositeForestSchema.__name__,
    dict_example = _composite_forest_example,
    default_file_path = "",
)



from __future__ import annotations

from dataclasses import dataclass, field
from typing import Union

from argrelay.composite_forest.CompositeNodeType import CompositeNodeType


@dataclass
class BaseNode:
    """
    Base class for all FS_33_76_82_84 composite forest node types.

    See also :class:`BaseNodeSchema`.
    """

    node_type: CompositeNodeType = field()
    """
    See `CompositeNodeType`.
    """

    sub_tree: Union[dict[str, BaseNode], None] = field()


@dataclass
class ZeroArgNode(BaseNode):
    """
    See FS_42_76_93_51 very first zero arg mapping interp.
    """

    plugin_instance_id: str = field()


@dataclass
class TreePathNode(BaseNode):
    """
    See `CompositeNodeType.tree_path_node`.
    """


@dataclass
class InterpTreeNode(BaseNode):
    """
    See FS_01_89_09_24 interp tree.
    """

    plugin_instance_id: str = field()


@dataclass
class FuncTreeNode(BaseNode):
    """
    See FS_26_43_73_72 func tree.
    """

    func_id: str = field()



from argrelay.client_pipeline.BytesHandlerAbstract import BytesHandlerAbstract
from argrelay.enum_desc.ClientExitCode import ClientExitCode
from argrelay.relay_client.proc_spinner import child_data_chunks
from argrelay.relay_client.proc_splitter import is_child_successful, get_child_exit_code


class BytesDstReceiver:

    def __init__(
        self,
        bytes_handler: BytesHandlerAbstract,
    ):
        super().__init__(
        )
        self.bytes_handler = bytes_handler

    def receive_bytes(
        self,
    ):
        if is_child_successful():
            self.handle_bytes(b"".join(child_data_chunks))
        else:
            client_exit_code = ClientExitCode(get_child_exit_code())
            exc_message = f"child exit_code: {client_exit_code} = {client_exit_code.name}"
            if client_exit_code is ClientExitCode.ConnectionError:
                raise ConnectionError(exc_message)
            else:
                raise RuntimeError(exc_message)

    def handle_bytes(
        self,
        given_bytes: bytes,
    ):
        self.bytes_handler.handle_bytes(given_bytes)



from argrelay.enum_desc.ClientExitCode import ClientExitCode
from argrelay.misc_helper_common import eprint
from argrelay.plugin_delegator.AbstractDelegator import AbstractDelegator
from argrelay.plugin_delegator.ErrorDelegatorCustomDataSchema import error_message_, error_code_
from argrelay.relay_server.LocalServer import LocalServer
from argrelay.runtime_context.InterpContext import InterpContext
from argrelay.schema_response.InvocationInput import InvocationInput


class ErrorDelegator(AbstractDelegator):

    def run_invoke_control(
        self,
        interp_ctx: InterpContext,
        local_server: LocalServer,
    ) -> InvocationInput:
        delegator_plugin_entry = local_server.plugin_config.plugin_instance_entries[
            self.plugin_instance_id
        ]
        invocation_input = InvocationInput.with_interp_context(
            interp_ctx,
            delegator_plugin_entry = delegator_plugin_entry,
            custom_plugin_data = {},
        )
        return invocation_input

    @staticmethod
    def invoke_action(invocation_input: InvocationInput):
        error_message = "ERROR: unknown error"
        error_code = ClientExitCode.GeneralError.value
        if invocation_input.custom_plugin_data:
            if error_message_ in invocation_input.custom_plugin_data:
                error_message = invocation_input.custom_plugin_data[error_message_]
            if error_code_ in invocation_input.custom_plugin_data:
                error_code = invocation_input.custom_plugin_data[error_code_]
        eprint(error_message)
        exit(error_code)
